#!/usr/bin/python
# -*- coding: utf-8 -*-
'''
KVBase (только строки) => KV (маршал) => OKV (объектные обёркти)

Скорость TC на записи строк:
=== tc ===
Запись 100000 строк: 0.11 s.
Чтение 100000 строк в случайном порядке: 0.14 s.
=== KVBase ===
Запись 100000 строк: 0.22 s.
Чтение 100000 строк в случайном порядке: 0.24 s.
=== KV ===
Запись 100000 строк: 0.33 s.
Чтение 100000 строк в случайном порядке: 0.32 s.
=== KVO ===
Запись 100000 строк: 0.33 s.
Чтение 100000 строк в случайном порядке: 0.64 s.
'''
import os
import marshal

def main():
    db = KVBase('var/test.tc', 'a+', 
        backup={'db_dir': 'var/backup/test.tc'}, compress=True)
    
    db['a'] = 'aaa'
    #~ db.restore_from_backup()
    #~ db.update_backup()
    print db['a']
    print db.backup_db['a']
    del db['a']
    
    return
    
    db = KVBase('var/test.tc', 'w')
    db['a'] = 'aaa'
    assert db['a'] == 'aaa'
    db.drop()

    db = KV('var/test.tc', 'w')
    db['a'] = {'a': 1}
    assert db['a']['a'] == 1
    db.drop()

    db = KVO('var/test.tc', 'w')
    db['a'] = {'a': 1}
    assert db['a'].a == 1
    db.drop()

    #~ speed_test(100000, 'var/test.tc')


def speed_test(row_count, db_fn):
    '''
    Тест скорости в режиме ключ-значение
    '''
    import random
    import om.t
    
    # данные для записи (ключ = значение)
    keys = [str(i) for i in xrange(row_count)]
    value = lambda key: key
    autocompress = 0

    for cls in [KVBase, UKVBase, KV, KVO]:
        print '=== %s ===' % cls.__name__
        
        # запись
        db = cls(db_fn, 'w', autocompress=autocompress)
        tm = om.t.T()
        for key in keys:
            db[key] = value(key)
        db.close()
        print 'Запись %s строк:' % len(keys), tm

        # чтение
        db = cls(db_fn, 'r', autocompress=autocompress)
        random.shuffle(keys)
        tm = om.t.T()
        for key in keys:
            val = db[key]
        db.close()
        print 'Чтение %s строк в случайном порядке:' % len(keys), tm
        print 'Размер базы:', db.size
        db.drop()


md = marshal.dumps
ml = marshal.loads


class KVBase(object):
    '''Работа со строками'''
    _opened = False
    
    def __init__(self, fn, mode='a+', db_type='tc', compress=False,
        autocompress=False, backup=None):
        '''
        *fn: файл
        **mode = режим открытия ('r', 'w', 'a', 'a+')
        **db_type: тип бд (tc, gdbm)
        **compress: сжимать данные
        **autocompress: сжимать данные если размер уменьшится
        **backup = {'db_dir': '/dir/name', **FSDBM_kwargs}
        '''
        self.fn = fn
        self.mode = mode
        self.db_type = db_type
        self._compress = compress
        if compress and db_type != 'tc':
            import zlib
            self.compress = lambda s: zlib.compress(s, 9)
            self.decompress = lambda s: zlib.decompress(s)
        elif autocompress:
            import zlib
            def compressor(s):
                compressed = zlib.compress(s, 9)
                if len(compressed) + 1 < len(s):
                    return '+' + compressed
                return '-' + s
            self.compress = compressor
            self.decompress = lambda s: (self.zlib.decompress(s[1:])
                if s.startswith('+') else s[1:])
        else:
            self.compress = lambda x: x
            self.decompress = lambda x: x
        self.backup = backup
        self.open()

    def open(self):
        dir = os.path.dirname(self.fn)
        if dir and not os.path.isdir(dir):
            os.makedirs(dir)
        if self.db_type == 'tc':
            import tc
            self.db = tc.HDB()
            if self._compress:
                self.db.tune(-1, -1, -1, tc.HDBTDEFLATE)
            self.db.open(self.fn, 
                {
                    'r': tc.HDBOREADER,
                    'w': tc.HDBOWRITER | tc.HDBOCREAT | tc.HDBOTRUNC,
                    'a': tc.HDBOWRITER,
                    'a+': tc.HDBOWRITER | tc.HDBOCREAT,
                }[self.mode]
            )
        elif self.db_type == 'gdbm':
            import gdbm
        else:
            assert 0, 'bad db type'
        if self.backup:
            import fsdbm
            self.backup_db = fsdbm.FSDBM(**self.backup)
        self._opened = True
    
    @property
    def size(self):
        return os.path.getsize(self.fn)

    def close(self):
        if self._opened:
            self.db.close()
        self._opened = False

    def drop(self):
        self.close()
        os.unlink(self.fn)
        if self.backup:
            self.backup_db.clear()
        
    def clear(self):
        self.drop()
        self.open()
    
    def __setitem__(self, key, value):
        self.db[key] = self.compress(value)
        if self.backup:
            self.backup_db[key] = value
        
    def __getitem__(self, key):
        return self.decompress(self.db[key])
        
    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default
            
    def __delitem__(self, key):
        del self.db[key]
        if self.backup:
            del self.backup_db[key]
        
    def __contains__(self, key):
        return key in self.db
        
    def __len__(self):
        return len(self.db)
        
    def __iter__(self):
        for k in self.db:
            yield k
    
    def iterkeys(self):
        return self.__iter__()
        
    def keys(self):
        return list(self)
        
    def itervalues(self):
        for k in self:
            yield self[k]

    def values(self):
        return list(self.itervalues())

    def iteritems(self):
        for k in self:
            yield (k, self[k])

    def items(self):
        return list(self.iteritems())

    def find(self, filter=None, order=None):
        '''
        Выборка документов
            **filter = lambda val: True|False
            **order = lambda val: True|False
        '''
        filter = filter or (lambda x: True)
        if order:
            items = [(k, order(v)) for k, v in self.iteritems() if filter(v)]
            items.sort(key=lambda x: x[1])
            return (self[k] for k, v in items)
        else:
            return (v for k, v in self.iteritems() if filter(v))

    def find_one(self, filter):
        for v in self.find(filter):
            return v
            
    def count(self, *args, **kwargs):
        '''
        Количество элементов по условию, параметры find
        '''
        i = 0
        for v in self.find(*args, **kwargs):
            i += 1
        return i
    
    def update_backup(self):
        '''Полностью обновить содержимое бэкапа'''
        if not self.backup:
            assert 0, 'backup db not found'
        self.backup_db.clear()
        for k, v in self.db.iteritems():
            self.backup_db[k] = v
    
    def restore_from_backup(self):
        '''Восстановить базу из бэкапа'''
        if not self.backup:
            assert 0, 'backup db not found'
        # clear db
        if self._opened:
            self.db.close()
        self._opened = False
        os.unlink(self.fn)
        self.open()
        for k, v in self.backup_db.iteritems():
            self.db[k] = v


class UKVBase(KVBase):
    '''Юникодные строки'''
    
    def __getitem__(self, key):
        return super(UKVBase, self).__getitem__(
            key.encode('utf-8')).decode('utf-8')
        
    def __setitem__(self, key, value):
        super(UKVBase, self).__setitem__(
            key.encode('utf-8'), value.encode('utf-8'))


class KV(KVBase):
    '''Маршализируемые данные'''
    
    def __getitem__(self, key):
        return ml(super(KV, self).__getitem__(key))
        
    def __setitem__(self, key, value):
        super(KV, self).__setitem__(key, md(value))


class ListDoc(list):
    def __init__(self, db, key, *args, **kwargs):
        super(ListDoc, self).__init__(*args, **kwargs)
        self._db = db
        self.key = key
        
    def save(self):
        self._db[self.key] = list(self)
        return self


class DictDoc(dict):
    def __init__(self, db, key, *args, **kwargs):
        super(DictDoc, self).__init__(*args, **kwargs)
        self.__dict__['_db'] = db
        self.__dict__['key'] = key
        
    def save(self):
        self._db[self.key] = dict(self)
        return self
        
    def __getattr__(self, name):
        return self[name]

    def __setattr__(self, name, val):
        self[name] = val

    def __delattr__(self, name):
        del self[name]


class KVObject(KV):
    '''Простенькие объектные обёртки на list и dict'''
    DOC_WRAPPERS = {dict: DictDoc, list: ListDoc}
    
    def doc_wrapper(self, key, obj):
        doc_cls = self.DOC_WRAPPERS.get(type(obj))
        return doc_cls(self, key, obj) if doc_cls else obj

    def __getitem__(self, key):
        ret = super(KVO, self).__getitem__(key)
        doc_cls = self.DOC_WRAPPERS.get(type(ret))
        if doc_cls:
            ret = doc_cls(self, key, ret)
        return ret
        


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
''' Полезные декораторы '''
import os
import sys
import time
import traceback


def main():
    import logging
    
    log = logging.getLogger(__name__)
    logging.basicConfig(level=logging.DEBUG, format='%(levelname)-8s %(message)s')

    @safely(default='foo', logger=log.error)
    @replay(3, logger=log.info)
    def test():
        pass
        1 / 0
    
    print test()


def cached_function(func):
    '''
    Кешируемая функция
    '''
    def _func(*args, **kwargs):
        try:
            return _func._retval
        except AttributeError:
            _func._retval = func(*args, **kwargs)
            return _func._retval
    return _func


class cached_property(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, instance, owner):
        value = self.f(instance)
        setattr(instance, self.f.__name__, value)
        return value


class cached_classmethod(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, instance, owner):
        value = self.f(owner)
        setattr(owner, self.f.__name__, value)
        return value


def safely(default=None, logger=None):
    '''
    Декоратор, игнорируем исключения
        **default = значение возвращаемое при ошибке
        **logger = функция для логгинга трейсбэков (например logging.error)
    '''
    def decorator(func):
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception, ex:
                if logger:
                    logger('SAFELY %s\n%s' %
                        (func.__name__, ''.join(traceback.format_exception(*sys.exc_info())[1:])))
                return default
        return wrapper
        wrapper.__name__ = func.__name__
    return decorator

def replay(cnt, sleep=0.1, logger=None):
    '''
    Декоратор, n-кратный перезапуск при исключении
        *cnt = сколько раз перезапускать
        **sleep = задержка между перезапусками, растёт с перезапусками
        **logger = функция для логгинга трейсбэков (например logging.debug)
    '''
    def decorator(func):
        def wrapper(*args, **kwargs):
            for i in xrange(cnt + 1):
                if i:
                    s = sleep * i
                    if logger:
                        logger('REPLAY %s (%i / %i), sleep %0.2f sec.\n%s' %
                            (func.__name__, i, cnt, s, ''.join(
                            traceback.format_exception(*sys.exc_info())[1:])))
                    if sleep:
                        time.sleep(s)
                try:
                    return func(*args, **kwargs)
                except Exception, e:
                    pass
            raise
        wrapper.__name__ = func.__name__
        return wrapper
    return decorator

def single_access(func=None, filename=None):
    '''
    Декоратор. Запускает, только если удаётся залочить файл.
    '''
    import file
    def decorator(func):
        lock_filename = filename or _func_loc_filename(func)
        def wrapper(*args, **kwargs):
            if file.lock(lock_filename):
                return func(*args, **kwargs)
            sys.stdout.write('Can\'t lock file: %s\n' % lock_filename)
        return wrapper
    if func:
        return decorator(func)
    else:
        return decorator

def _func_loc_filename(func):
    filename = sys.modules[func.__module__].__file__
    filename = os.path.abspath(filename)# + '.om.lock'
    return filename


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
''' Отправка сообщений '''
from decorators import safely

send = sys.modules[__name__]  # for auto-import

def main():
    from privat import GMAIL_LOGGER, MAIN_EMAIL

    all(GMAIL_LOGGER, MAIN_EMAIL, u'письмо', u'тема')
    #~ gmail(GMAIL_LOGGER[0], GMAIL_LOGGER[1], MAIN_EMAIL, u'письмо', u'тема')
    #~ gtalk(GMAIL_LOGGER[0], GMAIL_LOGGER[1], MAIN_EMAIL, u'сообщение')
    


def all(login_and_password, target, body, subject=''):
    login, passwd = login_and_password
    gmail(login, passwd, target, body, subject)
    gtalk(login, passwd, target,
        '[%s]\n%s' % (subject, body) if subject else body)
    


def gtalk(login, passwd, target, msg, safe=True):
    '''
    Отправка одиночного сообщения в jabber
        *jid        - jid отправителя
        *passwd     - пароль отправителя
        *target     - jid получателя
        *msg        - cообщение
        **safe      - (bool) игнорировать исключения
    '''
    import xmpp
    try:
        jid = xmpp.protocol.JID(login)
        cl = xmpp.Client(jid.getDomain(), debug=[])
        if not cl.connect(('talk.google.com', 5223)):
            raise IOError('Can not connect to server.')
        if not cl.auth(jid.getNode(), passwd):
            raise IOError('Can not auth with server.')
        cl.send(xmpp.Message(target, msg, typ='chat'))
        cl.disconnect()
    except (KeyboardInterrupt, SystemExit):
        raise
    except Exception, ex:
        if not safe:
            raise

def gmail(login, passwd, target, body, subject, safe=True):
    '''
    Отправка одиночного email через gmail
    '''
    try:
        g = Gmail(login, passwd)
        g.send(target, subject, body)
    except (KeyboardInterrupt, SystemExit):
        raise
    except Exception, ex:
        if not safe:
            raise

class Gmail(object):
    ''' Отправка мыл через gmail '''
    
    def __init__(self, sender, password):
        self.sender = str(sender)
        self.password = password
        self.server = None

    def auth(self):
        # аутентификация
        from smtplib import SMTP
        from socket import sslerror
        
        server = SMTP('smtp.gmail.com', 587)
        server.set_debuglevel(0)
        server.ehlo()
        server.starttls()
        server.ehlo()
        server.login(self.sender, self.password)
        self.server= server
        
    def send(self, to, subj, body):
        # отправка сообщения
        from smtplib import SMTP
        #~ from socket import sslerror
        if not self.server:
            self.auth()
        to, subj, body = [s.encode('utf-8') if isinstance(s, unicode) else s for s in (to, subj, body)]
        full_text = 'From: %s\r\nTo: %s\r\nSubject:%s\r\nContent-Type: text/plain; charset="utf-8"\r\n\r\n%s' % (self.sender, to, subj, body)
        rc = self.server.sendmail(self.sender, to, full_text)


if __name__ == "__main__":
    main()





# -*- coding: utf-8 -*-
try:
    import json
except ImportError:
    import simplejson as json


def pp(obj, *args, **kwargs):
    '''pprint с раскодированым выводом русских строк'''
    def default(v):
        try:
            ret = str(v)
        except Exception:
            ret = 'not json serializable'
        if len(ret) > 50:
            ret = ret[:50] + ' ...'
        return ret
    print json.dumps(obj, ensure_ascii=False, indent=4,
        default=default).encode('utf-8')


# -*- coding: utf-8 -*-
'''
om_thread.py - многопоточный генератор
'''
import sys
import threading
import time
import Queue
import traceback

thread = sys.modules[__name__]  # for auto-import
END = object()  # стоп-значение для потока


def main():
    '''
    Пример использования
    '''
    import logging
    from imbolc import om
    logging.basicConfig(level=logging.DEBUG, format='%(levelname)-8s %(message)s')
    
    def tst(i):
        1/0
        print om.url.load('zenpad.ru').body[:10]
        print '-> %i' % i
        time.sleep(0.1)
        return '<- %i' % i
    
    # получаем результаты по ходу вычислений
    for s in run(tst, xrange(10), 3, err_logger=logging.error):
        print s
        
    #~ # если результаты не нужны
    #~ list(run(tst, xrange(100), 3))
        

def run(callback, tasks, thread_limit, queue_limit=None, err_logger=None,
        sleep=0.01):
    '''
    Многопоточный генератор, на вход:
        callback        - рабочая функция (или запускаемый объект)
        tasks           - задачи для callback, любой итерируемый тип
        thread_limit    - количество потоков
        queue_limit     - максимальный размер очереди заданий
        err_logger      - логгер ошибок, например logging.error
        sleep           - время сна в цикле (для уменьшения загрузки процессора)
    '''
    queue = Queue.Queue(queue_limit)
    tasks = iter(tasks)
    pool = []
    ret_queue = Queue.Queue()
    queue_limit = queue_limit if queue_limit else thread_limit * 5
    
    while True:
        # возвращаем результат
        while not ret_queue.empty():
            yield ret_queue.get()
                
        # проверяем работоспособность потоков
        pool = filter(lambda x: x.isAlive(), pool)

        # добавляем недостающие потоки
        while len(pool) < thread_limit:
            th = Th(callback, queue, ret_queue, err_logger)
            th.setDaemon(True)
            th.start()
            pool.append(th)
            
        # заполняем очередь заданий
        while queue.qsize() < queue_limit:
            try:
                task = tasks.next()
            except StopIteration:
                break
            queue.put(task)
        else:
            # задания ещё не кончились
            time.sleep(sleep)
            continue
        break
        
    # отключаем потоки
    for th in pool:
        queue.put(END)

    # ожидаем завершения потоков
    for th in pool:
        th.join()
    
    # отдаём остатки
    while not ret_queue.empty():
        yield ret_queue.get()


class Th(threading.Thread):
    '''
    Поток
    '''
    def __init__(self, callback, queue, ret_queue=None, err_logger=None):
        self.callback = callback
        self.queue = queue
        self.ret_queue = ret_queue
        self.err_logger = err_logger
        threading.Thread.__init__(self)
        
    def run(self):
        while True:
            args = self.queue.get()
            if args is END:
                break
            try:
                ret = self.callback(args)
            except Exception:
                if self.err_logger:
                    self.err_logger(''.join(
                        traceback.format_exception(*sys.exc_info())[1:]))
                raise
            if self.ret_queue:
                self.ret_queue.put(ret)


if __name__ == "__main__":
    main()



import os


_proc_status = '/proc/%d/status' % os.getpid()

_scale = {
    'kB': 1024.0, 'mB': 1024.0 * 1024.0,
    'KB': 1024.0, 'MB': 1024.0 * 1024.0
}


def main():
    mt = MemoryTracker()
    a = 'a' * 1024 * 1024
    print mt


def _VmB(VmKey):
    '''Private.
    '''
    global _proc_status, _scale
     # get pseudo file  /proc/<pid>/status
    try:
        t = open(_proc_status)
        v = t.read()
        t.close()
    except:
        return 0.0  # non-Linux?
     # get VmKey line e.g. 'VmRSS:  9999  kB\n ...'
    i = v.index(VmKey)
    v = v[i:].split(None, 3)  # whitespace
    if len(v) < 3:
        return 0.0  # invalid format?
     # convert Vm value to bytes
    return float(v[1]) * _scale[v[2]]


def memory(since=0.0):
    '''Return memory usage in bytes.
    '''
    return _VmB('VmSize:') - since
    
def resident(since=0.0):
    '''Return resident memory usage in bytes.
    '''
    return _VmB('VmRSS:') - since


def stacksize(since=0.0):
    '''Return stack size in bytes.
    '''
    return _VmB('VmStk:') - since

    
class MemoryTracker(object):
    def __init__(self):
        self.start = resident()
    
    def b(self):
        return resident() - self.start

    def mb(self):
        return self.b() / 1024.0 / 1024

    def __str__(self):
        return '%0.2f MB' % (self.mb())


if __name__ == '__main__':
    main()